/****************************************************************************
 *
 * Palm Universal Wireless Keyboard(TM) driver with GTK1.2 v1.0
 * Copyright GPL Licence. 
 * 2006.01. 30 lampo1 <lampo1@paran.com>
 * Some code originated from "Teach yourself in 21days GTK+ prgramming" 
 * by Donna Martin et al. Sams Press.
 *
 ***************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <gtk/gtk.h>

#define VERSION v1.0
#define MAXNAME 16 

#define PRESSED  1
#define RELEASED 0
#define DONE     1
#define YET      0

#define GRAVE   0x0e
#define ONE     0x16
#define TWO     0x1e
#define THREE   0x26
#define FOUR    0x25
#define FIVE    0x2e
#define SIX     0x36
#define SEVEN   0x3d
#define EIGHT   0x3e
#define NINE    0x46
#define ZERO    0x45
#define MINUS   0x4e
#define EQUAL   0x55
#define BKSP    0x66
#define TAB     0x0d
#define Q       0x15
#define W       0x1d
#define E       0x24
#define R       0x2d
#define T       0x2c
#define Y       0x35
#define U       0x3c
#define I       0x43
#define O       0x44
#define P       0x4d
#define LBRCKT  0x54
#define RBRCKT  0x5b
#define BKSLSH  0x5d
#define CAPSLK  0x58
#define A       0x1c
#define S       0x1b
#define D       0x23
#define F       0x2b
#define G       0x34
#define H       0x33
#define J       0x3b
#define K       0x42
#define L       0x4b
#define SMCOLN  0x4c
#define APSTRP  0x52
#define ENTER   0x5a
#define LSHIFT  0x12
#define Z       0x1a
#define X       0x22
#define C       0x21
#define V       0x2a
#define B       0x32
#define N       0x31
#define M       0x3a
#define COMMA   0x48
#define PERIOD  0x49
#define UP      0x28
#define RSHIFT  0x59
#define SLASH   0x4a
#define LCTRL   0x14
#define LFN     0x02
#define CMD     0x03 
#define CANCEL  0x11
#define SPACE   0x29
#define ALTGR   0x30
#define LEFT    0x5e
#define DOWN    0x60
#define RIGHT   0x2f
#define DEL     0x1f

struct keymapping {
    unsigned char irbyte;

    char nml[MAXNAME];
    unsigned int kcode_nml;   /* keycode normal */

    char lfn[MAXNAME];
    unsigned int kcode_lfn;   /* keycode with LFN */
};  

struct keymapping KeyTable[64] = {
    {GRAVE,         "grave",0,           "",0},
    {ONE,               "1",0,           "",0},
    {TWO,               "2",0,           "",0},
    {THREE,             "3",0,           "",0},
    {FOUR,              "4",0,           "",0},
    {FIVE,              "5",0,           "",0},
    {SIX,               "6",0,           "",0},
    {SEVEN,             "7",0,           "",0},
    {EIGHT,             "8",0,           "",0},
    {NINE,              "9",0,           "",0},
    {ZERO,              "0",0,           "",0},
    {MINUS,         "minus",0,           "",0},
    {EQUAL,         "equal",0,           "",0},
    {BKSP,      "BackSpace",0,           "",0},
    {TAB,             "Tab",0,           "",0},
    {Q,                 "q",0,           "",0},
    {W,                 "w",0,           "",0},
    {E,                 "e",0,           "",0},
    {R,                 "r",0,           "",0},
    {T,                 "t",0,           "",0},
    {Y,                 "y",0,           "",0},
    {U,                 "u",0,           "",0},
    {I,                 "i",0,           "",0},
    {O,                 "o",0,           "",0},
    {P,                 "p",0,           "",0},
    {LBRCKT,  "bracketleft",0,           "",0},
    {RBRCKT, "bracketright",0,           "",0}, 
    {BKSLSH,    "backslash",0,           "",0},
    {CAPSLK,    "Caps_Lock",0,           "",0},
    {A,                 "a",0,           "",0},
    {S,                 "s",0,           "",0},
    {D,                 "d",0,           "",0},
    {F,                 "f",0,           "",0},
    {G,                 "g",0,           "",0},
    {H,                 "h",0,           "",0},
    {J,                 "j",0,           "",0},
    {K,                 "k",0,           "",0},
    {L,                 "l",0,           "",0},
    {SMCOLN,    "semicolon",0,           "",0},
    {APSTRP,   "apostrophe",0,           "",0},
    {ENTER,        "Return",0,           "",0},
    {LSHIFT,      "Shift_L",0,           "",0},
    {Z,                 "z",0,           "",0},
    {X,                 "x",0,           "",0},
    {C,                 "c",0,           "",0},
    {V,                 "v",0,           "",0},
    {B,                 "b",0,           "",0},
    {N,                 "n",0,           "",0},
    {M,                 "m",0,           "",0},
    {COMMA,         "comma",0,           "",0},
    {PERIOD,       "period",0,           "",0},
    {UP,               "Up",0,      "Prior",0},
    {RSHIFT,      "Shift_R",0,           "",0},
    {SLASH,         "slash",0,           "",0},
    {LCTRL,     "Control_L",0,           "",0},
    {CMD,          "Escape",0,           "",0}, 
    {SPACE,         "space",0,           "",0},
    {ALTGR,         "Alt_R",0,           "",0}, 
    {LEFT,           "Left",0,       "Home",0},
    {DOWN,           "Down",0,       "Next",0},
    {RIGHT,         "Right",0,        "End",0},
    {DEL,          "Delete",0,           "",0}, /* 62th */ 
    {CANCEL,             "",0,           "",0},  
    {LFN,                "",0,           "",0}, 
};


void exit_drv(int signum); 
void check_process_dup();
void read_keymap();
void read_rcfile(FILE *fp);
void translate(unsigned char ch);
void get_irbyte(); 
void gui(int argc, char **argv);
void cb_scalebar1 (GtkAdjustment *adj, GtkWidget *);
void cb_scalebar1 (GtkAdjustment *adj, GtkWidget *);
gboolean OnWindowDeleteEvent(GtkWidget* widget, GdkEventAny* event, gpointer data);
void OnWindowDestroy(GtkObject* object, gpointer data);

int ttyfd;                           /* file descriptor for ttySx*/
struct termios oldtio, newtio;      

unsigned int typematic_rate = 1;     /* n times of unit cycle */
unsigned int typematic_delay = 3;    /* n times of unit cycle */

Display *d;

void check_process_dup() {

    FILE *fp;
    char buf[255];

    fp = popen("ps -C uwkgtk1 -o pid= -o comm= | wc -l", "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot execute ps command.\n");
        exit(EXIT_FAILURE);
    }

    if ( atoi(fgets(buf, 255, fp)) >= 2 ) {
	fprintf(stderr, "uwkgtk1 is already running!\n");
	exit(EXIT_FAILURE);
    }

    if ( pclose(fp) == -1 ) {
	fprintf(stderr, "unable to close pipe.\n");
	exit(EXIT_FAILURE);
    }
}

void exit_drv(int signum) {

    if ( tcsetattr(ttyfd, TCSAFLUSH, &oldtio) == 0 ) {
	close(ttyfd);
        exit(EXIT_SUCCESS);
    } else { 
	close(ttyfd);
	exit(EXIT_FAILURE);
    }
}

void get_irbyte() {

    unsigned char ch, irbyte;
    int nread;

    nread = read(ttyfd, &ch, 1);
    if ( ( nread > 0 ) && ( ch == 0xff ) ) {    
	nread = read(ttyfd, &ch, 1);
	if ( (nread > 0) && ( ch == 0xc0 ) ) {	    
	    nread = read(ttyfd, &irbyte, 1);
	    read(ttyfd, &ch, 1);
	    read(ttyfd, &ch, 1); 
	    nread = read(ttyfd, &ch, 1);
	    if ( ch == 0xc1 ) {
		translate(irbyte);		
	    }
	} else if ( nread == -1 ) {
	    fprintf(stderr,"IrDA Port reading error.\n"
		    "Please test another serial port. Use -h option\n"
		    "Recommend removing other irda modules.\n"
		    );
	    exit(EXIT_FAILURE);
	}         
    }  
}


void cb_scalebar1 (GtkAdjustment *adj, GtkWidget *picture)
{
    typematic_delay = (unsigned int) adj->value;     
}

void cb_scalebar2 (GtkAdjustment *adj, GtkWidget *picture)
{
    typematic_rate = (unsigned int) adj->value; 

    #ifdef DEBUG
    printf("typematic_rate = %d\n", typematic_rate);
    #endif
}


gboolean OnWindowDeleteEvent(GtkWidget* widget, GdkEventAny* event, gpointer data) {

    /*return false so that window will be destroyed*/
    return FALSE;
}

void OnWindowDestroy(GtkObject* object, gpointer data) {

    /*exit main loop*/
    gtk_main_quit();
}

void SaveRcfile(GtkObject* object, gpointer data) {

    FILE *fp;
    
    fp = fopen("/etc/uwkrc", "w");
    if ( fp == NULL ) {
	printf("Error opening /etc/uwkrc\n");
	exit(EXIT_FAILURE);
    }
    
    fprintf( fp, 
	     "# rcfile for uwkgtk1 v1.0\n"
	     "delay until repeat : %d\n"
             "key repeat rate : %d\n", 
	     typematic_delay, typematic_rate
	   );
    
    if ( fclose(fp) == EOF) {
	printf("Error closing /etc/uwkrc\n");
	exit(EXIT_FAILURE);
    }
}
 

void gui(int argc, char** argv) {

    GtkWidget* wndw;			 

    GtkWidget* nb;	                      /* the notebook widget */
    GtkWidget* nb_pg1tab;
    GtkWidget* nb_pg2tab;
    GtkWidget* nb_pg3tab;
	
    GtkWidget* nb_pg1vbx;	       	      /* notebook page1 */

    GtkWidget* nb_pg1vbx_frm1;            /* frame1 : typematic delay */
    GtkWidget* nb_pg1vbx_frm1_hbx;
    GtkWidget* nb_pg1vbx_frm1_hbx_label1;   /* short */
    GtkObject* nb_pg1vbx_frm1_hbx_adj;
    GtkWidget* nb_pg1vbx_frm1_hbx_scale;
    GtkWidget* nb_pg1vbx_frm1_hbx_label2;   /* long */
    
    GtkWidget* nb_pg1vbx_frm2;            /* frame2 : typematic rate */
    GtkWidget* nb_pg1vbx_frm2_hbx;
    GtkWidget* nb_pg1vbx_frm2_hbx_label1;   /* slow */
    GtkObject* nb_pg1vbx_frm2_hbx_adj;
    GtkWidget* nb_pg1vbx_frm2_hbx_scale;
    GtkWidget* nb_pg1vbx_frm2_hbx_label2;   /* fast */

    GtkWidget* nb_pg1vbx_entry;

    GtkWidget* nb_pg1vbx_hbx;
    GtkWidget* nb_pg1vbx_hbx_btn1;        /* save */
    GtkWidget* nb_pg1vbx_hbx_btn2;        /* exit */


    GtkWidget* nb_pg2label;                 /* notebook page2 */
    GtkWidget* nb_pg3label;                 /* notebook page3 */
	
    /*init gtk+*/
    gtk_set_locale();
    gtk_init(&argc, &argv);
	
    /*create a window, set its title and hook up signal handlers (the usual stuff)*/
    wndw = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_policy(GTK_WINDOW(wndw), TRUE, TRUE, TRUE); 
    gtk_window_set_title(GTK_WINDOW(wndw), "UWKeyboard driver");
    gtk_window_set_default_size(GTK_WINDOW(wndw), 240, 280);
    gtk_signal_connect(GTK_OBJECT(wndw), "delete-event", GTK_SIGNAL_FUNC(OnWindowDeleteEvent), NULL);
    gtk_signal_connect(GTK_OBJECT(wndw), "destroy", GTK_SIGNAL_FUNC(OnWindowDestroy), NULL);
	
    /*create a notebook and pack it into the window*/
    nb = gtk_notebook_new();
    gtk_container_add(GTK_CONTAINER(wndw), nb);
	
    /*move page tabs to top of window*/
    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(nb), GTK_POS_TOP);
	
    /*make the tabs all be the same size*/
    gtk_notebook_set_homogeneous_tabs(GTK_NOTEBOOK(nb), TRUE);
	
    /*add some pages to the notebook*/

    nb_pg1tab = gtk_label_new("Settings");
    nb_pg1vbx = gtk_vbox_new(FALSE, 0);

    nb_pg1vbx_frm1 = gtk_frame_new(" Delay until repeat ");
    gtk_frame_set_label_align(GTK_FRAME(nb_pg1vbx_frm1), 0.0, 0.0);
    gtk_frame_set_shadow_type(GTK_FRAME(nb_pg1vbx_frm1), GTK_SHADOW_ETCHED_IN);

    nb_pg1vbx_frm1_hbx = gtk_hbox_new(FALSE,0);
    nb_pg1vbx_frm1_hbx_label1 = gtk_label_new("short");
    nb_pg1vbx_frm1_hbx_adj = gtk_adjustment_new( (float)typematic_delay, 0.0, 5.0, 1.0, 1.0, 1.0);
    nb_pg1vbx_frm1_hbx_scale = gtk_hscale_new(GTK_ADJUSTMENT(nb_pg1vbx_frm1_hbx_adj));
    gtk_range_set_update_policy( GTK_RANGE(nb_pg1vbx_frm1_hbx_scale), GTK_UPDATE_DISCONTINUOUS);
    gtk_scale_set_draw_value(GTK_SCALE(nb_pg1vbx_frm1_hbx_scale), FALSE);
    nb_pg1vbx_frm1_hbx_label2 = gtk_label_new("long");

    nb_pg1vbx_frm2 = gtk_frame_new(" Key repeat rate ");
    gtk_frame_set_label_align(GTK_FRAME(nb_pg1vbx_frm2), 0.0, 0.0);
    gtk_frame_set_shadow_type(GTK_FRAME(nb_pg1vbx_frm2), GTK_SHADOW_ETCHED_IN);
    nb_pg1vbx_frm2_hbx = gtk_hbox_new(FALSE,0);
    nb_pg1vbx_frm2_hbx_label1 = gtk_label_new("slow");
    nb_pg1vbx_frm2_hbx_adj = gtk_adjustment_new( (float)typematic_rate, 0.0, 5.0, 1.0, 1.0, 1.0);
    nb_pg1vbx_frm2_hbx_scale = gtk_hscale_new(GTK_ADJUSTMENT(nb_pg1vbx_frm2_hbx_adj));
    gtk_range_set_update_policy( GTK_RANGE(nb_pg1vbx_frm1_hbx_scale), GTK_UPDATE_DISCONTINUOUS);
    gtk_scale_set_draw_value(GTK_SCALE(nb_pg1vbx_frm2_hbx_scale), FALSE);
    nb_pg1vbx_frm2_hbx_label2 = gtk_label_new("fast");
	
    nb_pg1vbx_entry = gtk_entry_new();
    gtk_entry_set_editable(GTK_ENTRY(nb_pg1vbx_entry), TRUE);
    gtk_entry_set_text(GTK_ENTRY(nb_pg1vbx_entry), "Test here_");

    nb_pg1vbx_hbx = gtk_hbox_new(FALSE, 0);
    nb_pg1vbx_hbx_btn1 = gtk_button_new_with_label("Save");
    nb_pg1vbx_hbx_btn2 = gtk_button_new_with_label("Exit");

    gtk_box_pack_start_defaults(GTK_BOX(nb_pg1vbx_frm1_hbx), nb_pg1vbx_frm1_hbx_label1);	
    gtk_box_pack_start_defaults(GTK_BOX(nb_pg1vbx_frm1_hbx), nb_pg1vbx_frm1_hbx_scale);
    gtk_box_pack_start_defaults(GTK_BOX(nb_pg1vbx_frm1_hbx), nb_pg1vbx_frm1_hbx_label2);

    gtk_box_pack_start_defaults(GTK_BOX(nb_pg1vbx_frm2_hbx), nb_pg1vbx_frm2_hbx_label1);	
    gtk_box_pack_start_defaults(GTK_BOX(nb_pg1vbx_frm2_hbx), nb_pg1vbx_frm2_hbx_scale);
    gtk_box_pack_start_defaults(GTK_BOX(nb_pg1vbx_frm2_hbx), nb_pg1vbx_frm2_hbx_label2);

    gtk_container_add(GTK_CONTAINER(nb_pg1vbx_frm1), nb_pg1vbx_frm1_hbx);
    gtk_container_add(GTK_CONTAINER(nb_pg1vbx_frm2), nb_pg1vbx_frm2_hbx);

    gtk_box_pack_start(GTK_BOX(nb_pg1vbx_hbx), nb_pg1vbx_hbx_btn1, TRUE, TRUE, 20);
    gtk_box_pack_end(GTK_BOX(nb_pg1vbx_hbx), nb_pg1vbx_hbx_btn2, TRUE, TRUE, 20);

    gtk_box_pack_start(GTK_BOX(nb_pg1vbx), nb_pg1vbx_frm1, TRUE, TRUE, 10);
    gtk_box_pack_start(GTK_BOX(nb_pg1vbx), nb_pg1vbx_frm2, TRUE, TRUE, 10);  
    gtk_box_pack_start(GTK_BOX(nb_pg1vbx), nb_pg1vbx_entry, TRUE, TRUE, 10);
    gtk_box_pack_start(GTK_BOX(nb_pg1vbx), nb_pg1vbx_hbx, TRUE, TRUE,20);
		
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), nb_pg1vbx, nb_pg1tab);
    gtk_notebook_set_tab_label_packing(GTK_NOTEBOOK(nb), nb_pg1vbx, FALSE, FALSE, GTK_PACK_START);

    nb_pg2label = gtk_label_new("CMD : escape key\n"
			       "LSHIFT-SPACE : han/eng\n"			      
			       "LFN-R : reload keymap\n" 
			       "LFN-DEL : exit uwk driver\n"
                               "LCTRL-CMD : start menu\n"
                               "LCTRL-ALTGR-CMD : list\n"
                               "ALTGR-TAB : switch window\n"
			       "more key info : \n"
			       "http://www.icewm.org/manual\n"
			       );
    gtk_label_set_justify(GTK_LABEL(nb_pg2label), GTK_JUSTIFY_LEFT);
    nb_pg2tab = gtk_label_new("keys");
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), nb_pg2label, nb_pg2tab);

    nb_pg3label = gtk_label_new("Universal Wireless \n Keyboard(TM)\n"
			       "Yopy driver v2.0\n by lampo1@paran.com\n"
			       "2006. 1. 27.\n\n"
			       "http://irkbd.kldp.net");
    nb_pg3tab = gtk_label_new("Info");
    gtk_notebook_append_page(GTK_NOTEBOOK(nb), nb_pg3label, nb_pg3tab);	      


    gtk_signal_connect(GTK_OBJECT(nb_pg1vbx_hbx_btn1), "clicked", SaveRcfile, NULL);
    gtk_signal_connect(GTK_OBJECT(nb_pg1vbx_hbx_btn2), "clicked", OnWindowDestroy, NULL);
    gtk_signal_connect (GTK_OBJECT (nb_pg1vbx_frm1_hbx_adj), "value_changed",
                        GTK_SIGNAL_FUNC (cb_scalebar1), NULL);
    gtk_signal_connect (GTK_OBJECT (nb_pg1vbx_frm2_hbx_adj), "value_changed",
                        GTK_SIGNAL_FUNC (cb_scalebar2), NULL);

    gdk_input_add(ttyfd, GDK_INPUT_READ, get_irbyte, NULL); 

    /*show the window and all child widgets*/
    gtk_widget_show_all(wndw);
		
    /*enter main loop*/
    gtk_main();
}


int main(int argc, char** argv) {

    int option;
    FILE *fp;

    char *device = "/dev/ttyS1";
    char *rcfile = "/etc/uwkrc";

    struct sigaction newhandler;
    sigset_t blocked;

    check_process_dup();
    while ( (option = getopt(argc, argv, "hd:c:")) != -1 ) {
        switch(option) {
        case 'h' :
            fprintf(stderr,"Universal Wireless Keyboard(TM) GTK1 driver v1.0\n"
                           "Usage : uwkgtk1 [-d device] [-c rcfile]\n"
                           "Example : uwkgtk1 -d /dev/ttyS2 -c /etc/uwkrc\n"
                    );
            exit(EXIT_SUCCESS);
        case 'd' :
            device = optarg;
            break;
        case 'c' :
            rcfile = optarg;
            break;
        default :
            fprintf(stderr,"unknown option.\n");
            exit(EXIT_FAILURE);
        }
    }

    /* signal install */
    newhandler.sa_handler = exit_drv;
    sigemptyset(&blocked);
    sigaddset(&blocked, SIGALRM);
    newhandler.sa_mask = blocked;
    if ( sigaction(SIGALRM, &newhandler, NULL) == -1 ) {
        perror("sigaction install error\n");
        exit(EXIT_FAILURE);
    }

    /* open irda port device */
    ttyfd = open(device, O_RDONLY | O_NOCTTY );
    if (ttyfd == -1)  {
        fprintf(stderr, "Cannot open %s\n", device);
        exit(EXIT_FAILURE);
    }

    /* open rcfile */
    fp = fopen( rcfile, "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot open %s.\n"
                "So, default values are applied.\n", 
                rcfile
                );
    } else {
        read_rcfile( fp );
    }

    /* read current xmodmap */
    read_keymap();

   /* Serial port setting to noncanonical, no echo, ...*/
    tcgetattr(ttyfd, &oldtio); 
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;   /* inter-character timer unused */
    newtio.c_cc[VMIN]  = 1;   /* blocking read until 1 char received */
    tcsetattr(ttyfd,TCSAFLUSH,&newtio);
               
    d = XOpenDisplay(NULL);

    gui(argc, argv);
    return 0;
}


void read_keymap() {

    FILE *fp;
    int i;
    char buf[255];
    char cset[] = " \n";
    char *pname, *pcode;
 
    fp = popen("xmodmap -pke", "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot execute xmodmap -pke\n");
        exit(EXIT_FAILURE);
    }

    /* parsing the result of xmodmap -pke */
    while ( fgets(buf, 255, fp) ) {

	strtok(buf, cset);          /* discard "keycode" */
	pcode = strtok(NULL, cset); /* point to X keycode */
	strtok(NULL, cset);         /* discard "="  */
	pname = strtok(NULL, cset); /* point to X keyname */ 


	/* Compare a xmodmap line with UWK 64 keys.        */
	for ( i = 0 ; i <= 63; i++) {  
	    if ( pname != NULL ) {
		if ( !strcmp( KeyTable[i].nml, pname )) {
		    KeyTable[i].kcode_nml = atoi(pcode);
		} 
		if ( !strcmp( KeyTable[i].lfn, pname )) {
		    KeyTable[i].kcode_lfn = atoi(pcode);
		}
	    }
	}
    } 

    if ( pclose(fp) == -1 ) {
	fprintf(stderr, "unable to close pipe.\n");
	exit(EXIT_FAILURE);
    }

    #ifdef DEBUG
    printf("In printing...\n");
    for ( i = 0; i <= 63; i++ ) {
	printf("#%d %x %s %d %s %d\n",
	       i,
	       KeyTable[i].irbyte,
	       KeyTable[i].nml,
	       KeyTable[i].kcode_nml,
	       KeyTable[i].lfn,
	       KeyTable[i].kcode_lfn
	       );
    }
    #endif	       
}

void read_rcfile(FILE *fp) {

    char buf[255];    
    while ( (fgets(buf, 255, fp)) != NULL ) {

	if ( buf[0] == '#' ) continue;
	else if ( buf[0] == 'd' ) {
	    sscanf(buf, "%*s %d", &typematic_delay);
	} else if ( buf[0] == 'r' ) {
	    sscanf(buf, "%*s %d", &typematic_rate);
	} 
    }
    fclose(fp);
}


void translate( unsigned char ch ) {

    static int lfn_status = RELEASED;

/*
    static int sleep_status = YET;
    static int prev_ch   = 0 ;       
*/
    int  i;    
    unsigned char msb;            /* 0 : key pressed   */
                                  /* 1 : key released. */
    unsigned char bare_ch;

    msb =  (ch & 128) >> 7 ;   
    bare_ch = (ch & 127) ; 

/*  I'm thinking about typematic routines.      */
/*  Logically this code is correct and compact. */
/*  But, orginal UWK input timing is too slow.  */
/*  So, any delay routines only make driver's   */
/*  performance slow.                           */
/*  Therefore, I commented out.                 */
/*
    if ( ch == prev_ch ) {

	if (sleep_status == YET) {
	    sleep(typematic_delay);
	    sleep_status = DONE;
	    return;
	}
	sleep(typematic_rate);
		
    } else {
	sleep_status = YET;
	prev_ch = ch;
    }
*/	


    for (i = 0; i <= 63 ; i++) {

	if ( bare_ch == KeyTable[i].irbyte ) {
	    if ( bare_ch == LFN ) {              
		lfn_status = (1^msb);
		break;
	    } else if ( lfn_status == PRESSED ) {
		if ( bare_ch == DEL ) {          /* LFN-DEL */
		    close(ttyfd);
		    exit(EXIT_SUCCESS);
		}
		if ( bare_ch == EQUAL ) {        /* LFN-EQUAL */
		    if ( msb == 0 ) read_keymap();   /* read when pressed */
		    break;
		}

		XTestFakeKeyEvent(d, KeyTable[i].kcode_lfn, (1^msb), 0);
		XFlush(d);

                #ifdef DEBUG
		printf("code= %d\n", KeyTable[i].kcode_lfn);
                #endif


		break;		    
	    } else {
		XTestFakeKeyEvent(d, KeyTable[i].kcode_nml, (1^msb), 0);
		XFlush(d);

                #ifdef DEBUG
		printf("code= %d\n", KeyTable[i].kcode_nml);
                #endif

		break;
	    }
	}
    }
}
  


